/*  $Id: font_generic.cpp,v 1.32 2003/10/12 10:21:35 mbn Exp $
**
**  ClanLib Game SDK
**  Copyright (C) 2003  The ClanLib Team
**  For a total list of contributers see the file CREDITS.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "Display/display_precomp.h"
#include "font_generic.h"
#include "API/Core/Math/size.h"
#include "API/Core/Math/rect.h"
#include "API/Core/System/error.h"
#include "API/Core/System/clanstring.h"
#include "API/Display/sprite.h"
#include "API/Display/glyph_buffer.h"

#include <cmath>
#include <algorithm>
#include <vector>

using namespace std;

/////////////////////////////////////////////////////////////////////////////
// CL_Font_Generic construction:

CL_Font_Generic::CL_Font_Generic() :
	width_offset(0),
	height_offset(0),
	trans_origin(origin_top_left),
	trans_x(0),
	trans_y(0),
	delims(" "),
	fixed_width(0),	
	height(0),
	spacelen(0)
{
	set_letters("");
}

CL_Font_Generic::CL_Font_Generic(const CL_Font_Generic &copy) :
	width_offset(copy.width_offset),
	height_offset(copy.height_offset),
	trans_origin(copy.trans_origin),
	trans_x(copy.trans_x),
	trans_y(copy.trans_y),
	delims(copy.delims),
	fixed_width(copy.fixed_width),
	glyphs(copy.glyphs),
	height(copy.height),
	spacelen(copy.spacelen)
{
	cur_ctg = copy.cur_ctg;
	++(cur_ctg->second.ref_count);
}

CL_Font_Generic::~CL_Font_Generic()
{
	decrement();
}

/////////////////////////////////////////////////////////////////////////////
// CL_Font_Generic attributes:

//Static char-to-glyph-map cache
CL_Font_Generic::CTGVector_Cache CL_Font_Generic::ctg_vector_cache;

int CL_Font_Generic::get_width(unsigned char chr) const
{
	if (chr == '\n')
		return 0;
	
	int width = width_offset;
	
	if (fixed_width != 0)
	{
		width += fixed_width;
	}
	else
	{
		short glyph_num = cur_ctg->second.char_to_glyph_map[chr];
		if (glyph_num != -1)
			width += glyphs.get_frame_size(glyph_num).width;
		else
			width += spacelen;
	}
	
	float scale_x, scale_y;
	get_scale(scale_x, scale_y);
	return max(0, (int)(width*scale_x));
}

int CL_Font_Generic::get_height() const
{
	float scale_x, scale_y;
	get_scale(scale_x, scale_y);
	
	return (int)((height + height_offset) * scale_y);
}

bool CL_Font_Generic::is_glyph(unsigned char chr) const
{
	if (cur_ctg->second.char_to_glyph_map[chr] == -1)
		return false;
	
	return true;
}

/////////////////////////////////////////////////////////////////////////////
// CL_Font_Generic operations:

void CL_Font_Generic::draw_glyph(int x, int y, unsigned char chr, float ang) const
{
	short glyph_num = cur_ctg->second.char_to_glyph_map[chr];
	if (glyph_num != -1)
	{
		if (ang != 0.0)
			glyphs.rotate(ang);
		
		glyphs.set_frame(glyph_num);
		glyphs.draw(x, y);

		if (ang != 0.0)
			glyphs.rotate(-ang);
	}
}

void CL_Font_Generic::decrement()
{
	//Do not erase the blank cache entry, since it's used by default CL_Fonts
	if ((--(cur_ctg->second.ref_count) <= 0) && (cur_ctg->first != ""))
		ctg_vector_cache.cache.erase(cur_ctg);
}

CL_Font_Generic& CL_Font_Generic::operator =(const CL_Font_Generic &copy)
{
	decrement();
	cur_ctg = copy.cur_ctg;
	++(cur_ctg->second.ref_count);
	
	height = copy.height;
	width_offset = copy.width_offset;
	height_offset = copy.height_offset;
	spacelen = copy.spacelen;
	trans_origin = copy.trans_origin;
	trans_x = copy.trans_x;
	trans_y = copy.trans_y;
	glyphs = copy.glyphs;
	fixed_width = copy.fixed_width;
	delims = copy.delims;
	
	return *this;
}

void CL_Font_Generic::set_letters(const string &letters)
{
	if ((cur_ctg = ctg_vector_cache.cache.find(letters)) == ctg_vector_cache.cache.end())
	{
		cur_ctg = ctg_vector_cache.cache.insert(pair<string, CTGVector>(letters, CTGVector())).first;
		
		string::const_iterator it = letters.begin();
		for (unsigned short i=0; i < letters.length(); ++i)
		{
			cur_ctg->second.char_to_glyph_map[(unsigned char) *it] = i;
			++it;
		}
	}
	else
		++(cur_ctg->second.ref_count);
}

void CL_Font_Generic::create_font(
	const CL_Sprite &new_glyphs,
	const string &letters,
	int new_spacelen,
	bool monospace)
{
	//Size of char limit on most platforms is 256
	const int length = min((int)(letters.size()), 256);
	
	if ((length > new_glyphs.get_frame_count()) || (length == 0))
	{
		throw CL_Error(CL_String::format(
			"Font error: Letter characters: %1, Available font glyphs: %2", 
			length,
			new_glyphs.get_frame_count()
		));
	}
	
	//If monospace font requested, find the width of the widest glyph
	//Then set the fixed_width var to that width, which gwidth() checks for
	//Also set space to that width
	if (monospace)
	{
		for (short i=0; i < length; ++i)
		{
			int glyph_width = new_glyphs.get_frame_size(i).width;
			if (glyph_width > fixed_width)
				fixed_width = glyph_width;
		}
		
		new_spacelen = fixed_width;
	}
	//If not monospace, and space width not specified, then use average width as space width
	else if (new_spacelen < 0)
	{
		string::size_type space_pos = letters.find(' ');
		
		if (space_pos != string::npos) {
			//If there is a character for space, then use it
			new_spacelen = new_glyphs.get_frame_size((int)space_pos).width;
		} else {
			//Make the space size the average of all character sizes
			new_spacelen = 0;
			
			for (int pos = 0; pos < length; ++pos) {
				new_spacelen += new_glyphs.get_frame_size((int)(pos)).width;
			}
			
			new_spacelen /= length;
		}
	}
	
	for (short i=0; i < length; ++i)
	{
		int glyph_height = new_glyphs.get_frame_size(i).height;
		if (glyph_height > height)
			height = glyph_height;
	}
	
	glyphs = new_glyphs;
	glyphs.set_alignment(origin_top_left, 0, 0);
	glyphs.set_rotation_hotspot(origin_top_left, 0, 0);
	spacelen = new_spacelen;
	
	//If monospace font requested, find the width of the widest glyph
	//Then set the fixed_width var to that width, which gwidth() checks for
	if (monospace)
	{
		for (short i=0; i < length; ++i)
		{
			int glyph_width = new_glyphs.get_frame_size(i).width;
			if (glyph_width > fixed_width)
				fixed_width = glyph_width;
		}
	}	
	
	set_letters(letters);
}
