/*  $Id: surface_target_opengl.cpp,v 1.23 2003/12/17 15:25:37 mbn Exp $
**
**  ClanLib Game SDK
**  Copyright (C) 2003  The ClanLib Team
**  For a total list of contributers see the file CREDITS.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "Display/display_precomp.h"
#include "surface_target_opengl.h"
#include "gl_state_checker.h"
#include "graphic_context_opengl.h"
#ifdef WIN32
#include "WGL/display_window_opengl.h"
#else
#include "GLX/display_window_opengl.h"
#endif
#include "API/Display/pixel_buffer.h"
#include "API/Display/graphic_context.h"
#include "API/Display/pixel_format.h"
#include "API/Display/display.h"
#include "API/Core/System/error.h"
#include "API/GL/opengl.h"
#include "API/GL/opengl_state.h"
#include "API/GL/opengl_wrap.h"
#include <Hermes/H_Conv.h>

#include <GL/glu.h>

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_Target_OpenGL construction:

CL_Surface_Target_OpenGL::CL_Surface_Target_OpenGL(
	CL_PixelBuffer *provider,
	bool delete_provider,
	CL_Surface::Hint hint)
{
	CL_OpenGLState state(CL_Display::get_current_window()->get_gc());
	state.set_active();
	state.setup_2d();
	clEnable(CL_TEXTURE_2D);

	gc = 0;
#ifdef WIN32
	pbuffer_context = 0;
	pbuffer_dc = 0;
	pbuffer = 0;

	if (wglCreatePbufferARB == 0)
	{
		wglCreatePbufferARB = (ptr_wglCreatePbufferARB) wglGetProcAddress("wglCreatePbufferARB");
		wglCreatePbufferARB = (ptr_wglCreatePbufferARB) wglGetProcAddress("wglCreatePbufferARB");
		wglGetPbufferDCARB = (ptr_wglGetPbufferDCARB) wglGetProcAddress("wglGetPbufferDCARB");
		wglReleasePbufferDCARB = (ptr_wglReleasePbufferDCARB) wglGetProcAddress("wglReleasePbufferDCARB");
		wglDestroyPbufferARB = (ptr_wglDestroyPbufferARB) wglGetProcAddress("wglDestroyPbufferARB");
		wglQueryPbufferARB = (ptr_wglQueryPbufferARB) wglGetProcAddress("wglQueryPbufferARB");
	}
#else
	pbuffer = 0;
	pbuffer_context = 0;

	if (glXCreateGLXPbufferSGIX == 0)
	{
		glXCreateGLXPbufferSGIX = (ptr_glXCreateGLXPbufferSGIX) CL_OpenGL::get_proc_address("glXCreateGLXPbufferSGIX");
		glXDestroyGLXPbufferSGIX = (ptr_glXDestroyGLXPbufferSGIX) CL_OpenGL::get_proc_address("glXDestroyGLXPbufferSGIX");
		glXQueryGLXPbufferSGIX = (ptr_glXQueryGLXPbufferSGIX) CL_OpenGL::get_proc_address("glXQueryGLXPbufferSGIX");
		glXSelectEventSGIX = (ptr_glXSelectEventSGIX) CL_OpenGL::get_proc_address("glXSelectEventSGIX");
		glXGetSelectedEventSGIX = (ptr_glXGetSelectedEventSGIX) CL_OpenGL::get_proc_address("glXGetSelectedEventSGIX");

		glXChooseFBConfigSGIX = (ptr_glXChooseFBConfigSGIX) CL_OpenGL::get_proc_address("glXChooseFBConfigSGIX");
		glXCreateContextWithConfigSGIX = (ptr_glXCreateContextWithConfigSGIX) CL_OpenGL::get_proc_address("glXCreateContextWithConfigSGIX");
	}
#endif
	pbuffer_changed = false;
	texture_changed = true;

	// Convert provider to OpenGL texture format ABGR:
	CL_PixelFormat abgr_format;
	abgr_format.set_depth(32);
	abgr_format.enable_colorkey(false);
	abgr_format.set_alpha_mask(0xFF000000);
	abgr_format.set_blue_mask (0x00FF0000);
	abgr_format.set_green_mask(0x0000FF00);
	abgr_format.set_red_mask  (0x000000FF);

	provider->lock();

	screen_size.width = (float)provider->get_width();
	screen_size.height = (float)provider->get_height();
	
	texture_size.width = 1;
	texture_size.height = 1;
	while (texture_size.width < screen_size.width) texture_size.width *= 2;
	while (texture_size.height < screen_size.height) texture_size.height *= 2;

	CL_PixelBuffer *buffer = new CL_PixelBuffer(
		static_cast<int>(texture_size.width),
		static_cast<int>(texture_size.height),
		static_cast<int>(texture_size.width) * 4,
		abgr_format);

	provider->convert(
		buffer->get_data(), abgr_format,
		buffer->get_pitch(),
		CL_Rect(0, 0, static_cast<int>(screen_size.width), static_cast<int>(screen_size.height)),
		CL_Rect(0, 0, static_cast<int>(screen_size.width), static_cast<int>(screen_size.height)));

	provider->unlock();
	if (delete_provider) delete provider;

	buffer->lock();
	
	// Detect what level of alpha we require if application didnt tell us:
	if (hint == CL_Surface::hint_automatic)
	{
		bool found_colorkey = false;
		bool found_alpha = false;
	
		unsigned char *data = (unsigned char *) buffer->get_data();
		for (int y=0; y<screen_size.height; y++)
		{
			for (int x=0; x<screen_size.width; x++)
			{
				unsigned char alpha = data[x*4+3];
				if (alpha == 0) found_colorkey = true;
				else if (alpha != 255) found_alpha = true;
			}
			data += buffer->get_pitch();
		}
		
		if (found_alpha) hint = CL_Surface::hint_alpha;
		else if (found_colorkey) hint = CL_Surface::hint_colorkey;
		else hint = CL_Surface::hint_opaque;
	}
	
	// Figure out what is the optimal supported format for this card:
	CLint desired_format = CL_RGBA8;
	if (hint == CL_Surface::hint_opaque) desired_format = CL_RGB8;
	clTexImage2D(
		CL_PROXY_TEXTURE_2D,      // target
		0,                        // level
		desired_format,           // internalformat
		int(texture_size.width),  // width
		int(texture_size.height), // height
		0,                        // border
		CL_RGBA,                  // format
		CL_UNSIGNED_BYTE,         // type
		buffer->get_data());      // texels

	CLint obtained_format = desired_format;
	clGetTexLevelParameteriv(CL_PROXY_TEXTURE_2D, 0, CL_TEXTURE_INTERNAL_FORMAT, &obtained_format);
	switch (hint)
	{
	case CL_Surface::hint_alpha:
		break;
	
	case CL_Surface::hint_colorkey:
		// If we got 16 bit, try to get 5 bits for colors if possible.
		if (obtained_format == CL_RGBA4) desired_format = CL_RGB5_A1;
		break;
	
	case CL_Surface::hint_opaque:
		break;

	case CL_Surface::hint_automatic:
		break;
	}

	// Upload to OpenGL:
	clGenTextures(1, &handle);
	clBindTexture(CL_TEXTURE_2D, handle);

	clTexImage2D(
		CL_TEXTURE_2D,            // target
		0,                        // level
		desired_format,           // internalformat
		int(texture_size.width),  // width
		int(texture_size.height), // height
		0,                        // border
		CL_RGBA,                  // format
		CL_UNSIGNED_BYTE,         // type
		buffer->get_data());      // texels

	clTexParameteri(CL_TEXTURE_2D, CL_TEXTURE_MIN_FILTER, CL_LINEAR);
	clTexParameteri(CL_TEXTURE_2D, CL_TEXTURE_MAG_FILTER, CL_LINEAR);
	clTexParameteri(CL_TEXTURE_2D, CL_TEXTURE_WRAP_S, CL_CLAMP_TO_EDGE);
	clTexParameteri(CL_TEXTURE_2D, CL_TEXTURE_WRAP_T, CL_CLAMP_TO_EDGE);

	buffer->unlock();
	delete buffer;
}

CL_Surface_Target_OpenGL::~CL_Surface_Target_OpenGL()
{
#ifdef WIN32
	if (pbuffer_context)
	{
		wglMakeCurrent(0, 0);
		wglDeleteContext(pbuffer_context);
	}
	if (pbuffer_dc) wglReleasePbufferDCARB(pbuffer, pbuffer_dc);
	if (pbuffer) wglDestroyPbufferARB(pbuffer);
#else
	if (pbuffer_context)
	{
		glXMakeCurrent(CL_DisplayWindow_OpenGL::get_display(), None, 0);
		glXDestroyContext(CL_DisplayWindow_OpenGL::get_display(), pbuffer_context);
	}
	if (pbuffer) glXDestroyGLXPbufferSGIX(CL_DisplayWindow_OpenGL::get_display(), pbuffer);
#endif

	clDeleteTextures(1, &handle);
}

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_Target_OpenGL attributes:

int CL_Surface_Target_OpenGL::get_width() const
{
	return (int)screen_size.width;
}
	
int CL_Surface_Target_OpenGL::get_height() const
{
	return (int)screen_size.height;
}

CL_PixelBuffer CL_Surface_Target_OpenGL::get_pixeldata()
{
	return CL_PixelBuffer();
}
	
CL_GraphicContext *CL_Surface_Target_OpenGL::get_gc()
{
	if (gc == 0)
	{
		// Create working pbuffer:
#ifdef WIN32
		if (wglCreatePbufferARB == 0)
		{
			throw CL_Error("WGL_ARB_pbuffer OpenGL extension not supported by this card");
		}

		int attribList[1] = { 0 };

		PIXELFORMATDESCRIPTOR pfd =
		{
			sizeof(PIXELFORMATDESCRIPTOR),  // size of this pfd 
			1,                              // version number
			// PFD_DRAW_TO_WINDOW |            // support window
			PFD_SUPPORT_OPENGL //|            // support OpenGL
			//PFD_DOUBLEBUFFER |              // double buffered
			//PFD_DEPTH_DONTCARE
			,             // do you care about a zbuffer?
			PFD_TYPE_RGBA,                  // RGBA type
			24,                             // 24-bit color depth
			0, 0, 0, 0, 0, 0,               // color bits ignored
			8,                              // no alpha buffer
			0,                              // shift bit ignored
			0,                              // no accumulation buffer
			0, 0, 0, 0,                     // accum bits ignored
			0,                              // z-buffer
			0,                              // no stencil buffer
			0,                              // no auxiliary buffer
			PFD_MAIN_PLANE,                 // main layer
			0,                              // reserved
			0, 0, 0                         // layer masks ignored
		};

		int pixelformat = ChoosePixelFormat(wglGetCurrentDC(), &pfd);

//		int pixelformat = GetPixelFormat(wglGetCurrentDC());

		pbuffer = wglCreatePbufferARB(
			wglGetCurrentDC(),
			pixelformat,
			int(screen_size.width),
			int(screen_size.height),
			attribList);
		pbuffer_dc = wglGetPbufferDCARB(pbuffer);
		pbuffer_context = wglCreateContext(pbuffer_dc);

		if (!CL_DisplayWindow_OpenGL::opengl_contexts.empty())
		{
			wglShareLists(CL_DisplayWindow_OpenGL::opengl_contexts.back(), pbuffer_context);
		}
#else
		if (glXCreateGLXPbufferSGIX == 0)
		{
			throw CL_Error("GLX_SGIX_pbuffer OpenGL extension not supported by this card");
		}

		int configAttribs[] =
		{
//			GLX_DOUBLEBUFFER,        False,
			GLX_RED_SIZE,            8,
			GLX_GREEN_SIZE,          8,
			GLX_BLUE_SIZE,           8,
			GLX_ALPHA_SIZE,          8,
			GLX_DRAWABLE_TYPE_SGIX,  GLX_PBUFFER_BIT_SGIX,
			(int) None
		};

		int num_configs_returned = 0;
		GLXFBConfig *config = glXChooseFBConfigSGIX(
			CL_DisplayWindow_OpenGL::get_display(),
			DefaultScreen(CL_DisplayWindow_OpenGL::get_display()),
			configAttribs,
			&num_configs_returned);

		pbuffer = glXCreateGLXPbufferSGIX(
			CL_DisplayWindow_OpenGL::get_display(),
			*config,
			int(screen_size.width),
			int(screen_size.height),
			0);

		pbuffer_context = glXCreateContextWithConfigSGIX(
			CL_DisplayWindow_OpenGL::get_display(),
			config[0],
			GLX_RGBA_TYPE,
			CL_DisplayWindow_OpenGL::share_context,
			True);
#endif

		// Create GC interface for it:
		gc = new CL_GraphicContext(new CL_GraphicContext_OpenGL(this));
		gc->set_cliprect(CL_Rect(0,0,int(screen_size.width),int(screen_size.height)));

		// Download surface to pbuffer:
		pbuffer_changed = false;
		texture_changed = true;
		sync_pbuffer();
	}

	return gc;
}

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_Target_OpenGL operations:

void CL_Surface_Target_OpenGL::set_pixeldata(
	const CL_Point &pos,
	const CL_Rect &src_rect,
	CL_PixelBuffer &pixeldata)
{
	CL_GLStateChecker::from_gc(CL_Display::get_current_window()->get_gc())->flush();

	// Go look if OpenGL already know this format:
	const CL_PixelFormat &src_format = pixeldata.get_format();
	
	int gl_format = CL_RGBA8;
	int gl_type = CL_UNSIGNED_BYTE;
	bool do_pb_convert = false;
	if (src_format == CL_PixelFormat::rgba8888)
	{
		gl_format = CL_RGBA;
		gl_type = CL_UNSIGNED_INT_8_8_8_8;
	}
	else if (src_format == CL_PixelFormat::rgb888)
	{
		gl_format = CL_BGR;
		gl_type = CL_UNSIGNED_BYTE;
	}
	else if (src_format == CL_PixelFormat::rgba4444)
	{
		gl_format = CL_RGBA;
		gl_type = CL_UNSIGNED_SHORT_4_4_4_4;
	}
	else if (src_format == CL_PixelFormat::rgb565)
	{
		gl_format = CL_RGB;
		gl_type = CL_UNSIGNED_SHORT_5_6_5;
	}
	else if (src_format == CL_PixelFormat::abgr8888)
	{
		gl_format = CL_RGBA;
		gl_type = CL_UNSIGNED_INT_8_8_8_8_REV;
	}
	else if (src_format == CL_PixelFormat::bgr888)
	{
		gl_format = CL_RGB;
		gl_type = CL_UNSIGNED_BYTE;
	}
	else if (src_format == CL_PixelFormat::abgr4444)
	{
		gl_format = CL_RGBA;
		gl_type = CL_UNSIGNED_SHORT_4_4_4_4_REV;
	}
	else if (src_format == CL_PixelFormat::bgr555)
	{
		gl_format = CL_RGB;
		gl_type = CL_UNSIGNED_SHORT_5_6_5_REV;
	}
	else
	{
		// Unsupported format by opengl
		gl_format = CL_RGBA;
		gl_type = CL_UNSIGNED_INT_8_8_8_8;
		do_pb_convert = true;
	}

	pixeldata.lock();

	if (do_pb_convert)
	{
		CL_PixelBuffer buffer(
			src_rect.get_width(),
			src_rect.get_height(),
			src_rect.get_width() * 4,
			CL_PixelFormat::rgba8888);

		buffer.lock();

		pixeldata.convert(
			buffer.get_data(),
			CL_PixelFormat::rgba8888,
			buffer.get_pitch(),
			CL_Rect(0, 0, src_rect.get_width(), src_rect.get_height()),
			src_rect);

		clTexSubImage2D(
			CL_TEXTURE_2D,
			0,
			pos.x,
			pos.y,
			src_rect.get_width(),
			src_rect.get_height(),
			gl_format,
			gl_type,
			buffer.get_data());

		buffer.unlock();
	}
	else
	{
		if (src_format.get_depth()*pixeldata.get_width()/8 != (int)pixeldata.get_pitch())
		{
			clPixelStorei(CL_UNPACK_ALIGNMENT, pixeldata.get_pitch());
		}
		clPixelStorei(CL_UNPACK_ROW_LENGTH, pixeldata.get_width());
		clPixelStorei(CL_UNPACK_SKIP_PIXELS, src_rect.left);
		clPixelStorei(CL_UNPACK_SKIP_ROWS, src_rect.top);

		clTexSubImage2D(
			CL_TEXTURE_2D,
			0,
			pos.x,
			pos.y,
			src_rect.get_width(),
			src_rect.get_height(),
			gl_format,
			gl_type,
			pixeldata.get_data());

		if (src_format.get_depth()*pixeldata.get_width()/8 != (int)pixeldata.get_pitch())
		{
			clPixelStorei(CL_UNPACK_ALIGNMENT, 0);
		}
		clPixelStorei(CL_UNPACK_ROW_LENGTH, -1);
		clPixelStorei(CL_UNPACK_SKIP_PIXELS, 0);
		clPixelStorei(CL_UNPACK_SKIP_ROWS, 0);
	}

	pixeldata.unlock();

	sync_pbuffer();
}

//: Draw surface on screen.
void CL_Surface_Target_OpenGL::draw(
	CL_Surface_Generic *attributes,
	int x, int y,
	CL_GraphicContext *gc)
{
	sync_texture();

	// Calculate translation hotspot
	CL_Pointf translation = calc_hotspot(
		attributes->translation_origin,
		attributes->translation_hotspot,
		screen_size);

	init_attributes(attributes, screen_size);

	if (attributes->scale_x == 1.0f && attributes->scale_y == 1.0f)
	{
		float x_pos = x - translation.x;
		float y_pos = y - translation.y;
		state.rotation_hotspot.x += x_pos;
		state.rotation_hotspot.y += y_pos;

		CL_GLStateChecker::from_gc(gc)->add_quad(
			state,
			x_pos,
			y_pos,
			x_pos + static_cast<int>(screen_size.width),
			y_pos + static_cast<int>(screen_size.height),
			0,
			0,
			static_cast<int>(screen_size.width),
			static_cast<int>(screen_size.height));
	}
	else
	{
		float x_pos = x - translation.x * attributes->scale_x;
		float y_pos = y - translation.y * attributes->scale_y;
		state.rotation_hotspot.x = x_pos + state.rotation_hotspot.x * attributes->scale_x;
		state.rotation_hotspot.y = y_pos + state.rotation_hotspot.y * attributes->scale_y;

		CL_GLStateChecker::from_gc(gc)->add_quad(
			state,
			x_pos,
			y_pos,
			x_pos + screen_size.width * attributes->scale_x,
			y_pos + screen_size.height * attributes->scale_y,
			0,
			0,
			static_cast<int>(screen_size.width),
			static_cast<int>(screen_size.height));
	}
}

//: Draw surface on screen.
void CL_Surface_Target_OpenGL::draw(
	CL_Surface_Generic *attributes,
	const CL_Rect &dest,
	CL_GraphicContext *gc)
{
	sync_texture();

	CL_Sizef size((float) dest.get_width(), (float) dest.get_height());
	
	// Calculate translation hotspot
	CL_Pointf translation = calc_hotspot(
		attributes->translation_origin,
		attributes->translation_hotspot,
		size);

//	CL_GLState state = init_attributes(attributes, size);
	init_attributes(attributes, size);

	if (attributes->scale_x == 1.0f && attributes->scale_y == 1.0f)
	{
		state.rotation_hotspot.x += dest.left-translation.x;
		state.rotation_hotspot.y += dest.top-translation.y;

		CL_GLStateChecker::from_gc(gc)->add_quad(
			state,
			dest.left-translation.x,
			dest.top-translation.y,
			dest.right-translation.x,
			dest.bottom-translation.y,
			0,
			0,
			dest.get_width(),
			dest.get_height());
	}
	else
	{
		translation.x *= attributes->scale_x;
		translation.y *= attributes->scale_y;

		state.rotation_hotspot.x = dest.left-translation.x + state.rotation_hotspot.x * attributes->scale_x;
		state.rotation_hotspot.y = dest.top-translation.y + state.rotation_hotspot.y * attributes->scale_y;

		CL_GLStateChecker::from_gc(gc)->add_quad(
			state,
			dest.left-translation.x,
			dest.top-translation.y,
			dest.left+(dest.right-dest.left)*attributes->scale_x-translation.x,
			dest.top+(dest.bottom-dest.top)*attributes->scale_y-translation.y,
			0,
			0,
			dest.get_width(),
			dest.get_height());
	}
}

//: Draw surface on screen.
void CL_Surface_Target_OpenGL::draw(
	CL_Surface_Generic *attributes,
	const CL_Rect &src,
	const CL_Rect &dest,
	CL_GraphicContext *gc)
{
	sync_texture();

	CL_Sizef size((float)dest.get_width(), (float)dest.get_height());

	// Calculate translation hotspot
	CL_Pointf translation = calc_hotspot(
		attributes->translation_origin,
		attributes->translation_hotspot,
		size);

	init_attributes(attributes, size);

	if (attributes->scale_x == 1.0f && attributes->scale_y == 1.0f)
	{
		state.rotation_hotspot.x += dest.left-translation.x;
		state.rotation_hotspot.y += dest.top-translation.y;

		CL_GLStateChecker::from_gc(gc)->add_quad(
			state,
			dest.left-translation.x,
			dest.top-translation.y,
			dest.right-translation.x,
			dest.bottom-translation.y,
			src.left,
			src.top,
			src.right,
			src.bottom);
	}
	else
	{
		translation.x *= attributes->scale_x;
		translation.y *= attributes->scale_y;

		state.rotation_hotspot.x = dest.left-translation.x + state.rotation_hotspot.x * attributes->scale_x;
		state.rotation_hotspot.y = dest.top-translation.y + state.rotation_hotspot.y * attributes->scale_y;

		CL_GLStateChecker::from_gc(gc)->add_quad(
			state,
			dest.left-translation.x,
			dest.top-translation.y,
			dest.left+(dest.right-dest.left)*attributes->scale_x-translation.x,
			dest.top+(dest.bottom-dest.top)*attributes->scale_y-translation.y,
			src.left,
			src.top,
			src.right,
			src.bottom);
	}
}

void CL_Surface_Target_OpenGL::sync_pbuffer()
{
	if (texture_changed && gc != 0)
	{
		CL_OpenGLState state(gc);
		state.set_active();

		// Setup down/up 2d projection matrix:
		clMatrixMode(CL_PROJECTION);
		clLoadIdentity();

		CLdouble viewport[4];
		clGetDoublev(CL_VIEWPORT, viewport);

		CLdouble width = viewport[2];
		CLdouble height = viewport[3];
		gluOrtho2D(0.0, width, 0.0, height);

		clMatrixMode(CL_MODELVIEW);
		clLoadIdentity();

		// Copy texture to pbuffer:
		clEnable(CL_TEXTURE_2D);
		clBindTexture(CL_TEXTURE_2D, handle);
		clMatrixMode(CL_TEXTURE);
		CLdouble scale_matrix[16];
		memset(scale_matrix, 0, sizeof(CLdouble)*16);
		scale_matrix[0] = 1.0/texture_size.width;
		scale_matrix[5] = 1.0/texture_size.height;
		scale_matrix[10] = 1.0;
		scale_matrix[15] = 1.0;
		clLoadMatrixd(scale_matrix);
		clMatrixMode(CL_MODELVIEW);

		clBegin(CL_QUADS);
		clColor3f(1.0f, 1.0f, 1.0f);
		clTexCoord2f(screen_size.width, 0.0f);
		clVertex2f(screen_size.width, 0.0f);
		clTexCoord2f(screen_size.width, screen_size.height);
		clVertex2f(screen_size.width, screen_size.height);
		clTexCoord2f(0.0f, screen_size.height);
		clVertex2f(0.0f, screen_size.height);
		clTexCoord2f(0.0f, 0.0f);
		clVertex2f(0.0f, 0.0f);
		clEnd();
	}
	texture_changed = false;
}

void CL_Surface_Target_OpenGL::sync_texture()
{
	if (pbuffer_changed && gc != 0)
	{
		CL_OpenGLState state(gc);
		state.set_active();

		// Copy pbuffer to texture:
		clEnable(CL_TEXTURE_2D);
		clBindTexture(CL_TEXTURE_2D, handle);
		clCopyTexSubImage2D(
			CL_TEXTURE_2D,
			0,
			0, 0,
			0, 0,
			int(screen_size.width), int(screen_size.height));
	}
	pbuffer_changed = false;
}

void CL_Surface_Target_OpenGL::set_pbuffer_modified()
{
	pbuffer_changed = true;
}

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_Target_OpenGL implementation:

#ifdef WIN32
CL_Surface_Target_OpenGL::ptr_wglCreatePbufferARB CL_Surface_Target_OpenGL::wglCreatePbufferARB = 0;

CL_Surface_Target_OpenGL::ptr_wglGetPbufferDCARB CL_Surface_Target_OpenGL::wglGetPbufferDCARB = 0;

CL_Surface_Target_OpenGL::ptr_wglReleasePbufferDCARB CL_Surface_Target_OpenGL::wglReleasePbufferDCARB = 0;

CL_Surface_Target_OpenGL::ptr_wglDestroyPbufferARB CL_Surface_Target_OpenGL::wglDestroyPbufferARB = 0;

CL_Surface_Target_OpenGL::ptr_wglQueryPbufferARB CL_Surface_Target_OpenGL::wglQueryPbufferARB = 0;
#else
CL_Surface_Target_OpenGL::ptr_glXCreateGLXPbufferSGIX CL_Surface_Target_OpenGL::glXCreateGLXPbufferSGIX = 0;

CL_Surface_Target_OpenGL::ptr_glXDestroyGLXPbufferSGIX CL_Surface_Target_OpenGL::glXDestroyGLXPbufferSGIX = 0;

CL_Surface_Target_OpenGL::ptr_glXQueryGLXPbufferSGIX CL_Surface_Target_OpenGL::glXQueryGLXPbufferSGIX = 0;

CL_Surface_Target_OpenGL::ptr_glXSelectEventSGIX CL_Surface_Target_OpenGL::glXSelectEventSGIX = 0;

CL_Surface_Target_OpenGL::ptr_glXGetSelectedEventSGIX CL_Surface_Target_OpenGL::glXGetSelectedEventSGIX = 0;

CL_Surface_Target_OpenGL::ptr_glXChooseFBConfigSGIX CL_Surface_Target_OpenGL::glXChooseFBConfigSGIX = 0;
	
CL_Surface_Target_OpenGL::ptr_glXCreateContextWithConfigSGIX CL_Surface_Target_OpenGL::glXCreateContextWithConfigSGIX = 0;
#endif
